import unittest

from toripchanger import server


class TestTorIpChangerServer(unittest.TestCase):
    def setUp(self):
        self.tor_ip_changer = unittest.mock.Mock()
        app = server.init_server(self.tor_ip_changer)
        app.testing = True

        self.client = app.test_client()

    def test_index(self):
        """
        Test `/` endpoint returns a simple text message.
        """
        r = self.client.get("/")
        self.assertEqual(r.status_code, 200)
        self.assertEqual(r.data.decode("utf-8"), "TorIpChanger Server")

    def test_changeip(self):
        """
        Test `/changeip/` endpoint is capable of changing Tor' IP.
        """
        self.tor_ip_changer.get_new_ip.return_value = "1.2.3.4"
        r = self.client.get("/changeip/")

        self.assertEqual(r.status_code, 200)
        self.assertEqual(r.json, {"newIp": "1.2.3.4", "error": ""})

    def test_changeip_exception(self):
        """
        Test `/changeip/` gracefully handles all exceptions.
        """
        self.tor_ip_changer.get_new_ip.side_effect = [ValueError("message")]
        r = self.client.get("/changeip/")

        self.assertEqual(r.status_code, 500)
        self.assertEqual(
            r.json, {"error": "<class 'ValueError'>: message", "newIp": ""}
        )



import ipaddress
from time import sleep

from requests import get
from requests.exceptions import RequestException
from stem import Signal
from stem.control import Controller

from toripchanger.exceptions import TorIpError

# Default settings.
REUSE_THRESHOLD = 1
LOCAL_HTTP_PROXY = "127.0.0.1:8118"
NEW_IP_MAX_ATTEMPTS = 10
TOR_PASSWORD = ""
TOR_ADDRESS = "127.0.0.1"
TOR_PORT = 9051
POST_NEW_IP_SLEEP = 0.5


# Service to get current IP.
ICANHAZIP = "http://icanhazip.com/"


class TorIpChanger:
    def __init__(
        self,
        reuse_threshold=REUSE_THRESHOLD,
        local_http_proxy=LOCAL_HTTP_PROXY,
        tor_password=TOR_PASSWORD,
        tor_address=TOR_ADDRESS,
        tor_port=TOR_PORT,
        new_ip_max_attempts=NEW_IP_MAX_ATTEMPTS,
        post_new_ip_sleep=POST_NEW_IP_SLEEP,
    ):
        """
        TorIpChanger - make sure requesting a new Tor IP address really does
        return a new (different) IP.

        The 'reuse_threshold' argument specifies the number of unique Tor IPs
        kept in the 'used_ips' list which works as a FIFO queue.

        When 'reuse_threshold' is set to 0 then none of the already used Tor IP
        addresses can be reused.
        When 'reuse_threshold' is set to 1 (which is default) then the current
        Tor IP address can be reused after one other IP was used.
        If 'reuse_threshold' is set to, for example, 5 then the current Tor IP
        address can be reused after 5 other Tor IPs were used.

        :argument reuse_threshold: IPs to use before reusing the current one
        :type reuse_threshold: int
        :argument local_http_proxy: local proxy IP and port
        :type local_http_proxy: str
        :argument tor_password: Tor password
        :type tor_password: str
        :argument tor_address: IP address of the Tor controller
        :type tor_address: str
        :argument tor_port: port number of the Tor controller
        :type tor_port: int
        :argument new_ip_max_attempts: get new IP attemps limit
        :type new_ip_max_attempts: int
        :argument post_new_ip_sleep: how long to wait after requesting a new IP
        :type post_new_ip_sleep: float
        """
        self.reuse_threshold = reuse_threshold
        self.local_http_proxy = local_http_proxy
        self.tor_password = tor_password
        self.tor_address = tor_address
        self.tor_port = tor_port
        self.new_ip_max_attempts = new_ip_max_attempts
        self.post_new_ip_sleep = post_new_ip_sleep

        self.used_ips = []  # We cannot use set() because order matters.

    @property
    def real_ip(self):
        """
        The actual public IP of this host.
        """
        if not hasattr(self, "_real_ip"):
            response = get(ICANHAZIP)
            self._real_ip = self._get_response_text(response)

        return self._real_ip

    def get_current_ip(self):
        """
        Get the current IP Tor is using.

        :returns str
        :raises TorIpError
        """
        response = get(ICANHAZIP, proxies={"http": self.local_http_proxy})

        if response.ok:
            return self._get_response_text(response)

        raise TorIpError("Failed to get the current Tor IP")

    def get_new_ip(self):
        """
        Try to obtain new a usable TOR IP.

        :returns bool
        :raises TorIpError
        """
        attempts = 0

        while True:
            if attempts == self.new_ip_max_attempts:
                raise TorIpError("Failed to obtain a new usable Tor IP")

            attempts += 1

            try:
                current_ip = self.get_current_ip()
            except (RequestException, TorIpError):
                self._obtain_new_ip()
                continue

            if not self._ip_is_usable(current_ip):
                self._obtain_new_ip()
                continue

            self._manage_used_ips(current_ip)
            break

        return current_ip

    def _get_response_text(self, response):
        return response.text.strip()

    def _ip_is_safe(self, current_ip):
        """
        Check if it's safe to (re-)use the current IP.

        :argument current_ip: current Tor IP
        :type current_ip: str

        :returns bool
        """
        return current_ip not in self.used_ips

    def _ip_is_usable(self, current_ip):
        """
        Check if the current Tor's IP is usable.

        :argument current_ip: current Tor IP
        :type current_ip: str

        :returns bool
        """
        # Consider IP addresses only.
        try:
            ipaddress.ip_address(current_ip)
        except ValueError:
            return False

        # Never use real IP.
        if current_ip == self.real_ip:
            return False

        # Do dot allow IP reuse.
        if not self._ip_is_safe(current_ip):
            return False

        return True

    def _manage_used_ips(self, current_ip):
        """
        Handle registering and releasing used Tor IPs.

        :argument current_ip: current Tor IP
        :type current_ip: str
        """
        # Register current IP.
        self.used_ips.append(current_ip)

        # Release the oldest registred IP.
        if self.reuse_threshold:
            if len(self.used_ips) > self.reuse_threshold:
                del self.used_ips[0]

    def _obtain_new_ip(self):
        """
        Change Tor's IP.
        """
        with Controller.from_port(
            address=self.tor_address, port=self.tor_port
        ) as controller:
            controller.authenticate(password=self.tor_password)
            controller.signal(Signal.NEWNYM)

        # Wait till the IP 'settles in'.
        sleep(self.post_new_ip_sleep)



import unittest
from unittest.mock import patch, Mock

from stem import Signal

from toripchanger.changer import (
    ICANHAZIP,
    LOCAL_HTTP_PROXY,
    NEW_IP_MAX_ATTEMPTS,
    TorIpChanger,
    TorIpError,
    TOR_PASSWORD,
    TOR_ADDRESS,
    TOR_PORT,
)


class TestTorIpChanger(unittest.TestCase):
    def test_init(self):
        """
        Test that 'TorIpChanger' init sets expected (default) attributes.
        """
        tor_ip_changer = TorIpChanger()

        self.assertEqual(tor_ip_changer.reuse_threshold, 1)
        self.assertFalse(tor_ip_changer.used_ips)

    def test_init_reuse_threshold(self):
        """
        Test that 'TorIpChanger' init sets reuse_threshold when proxied.
        """
        tor_ip_changer = TorIpChanger(5)

        self.assertEqual(tor_ip_changer.reuse_threshold, 5)

    @patch("toripchanger.changer.get")
    @patch("toripchanger.changer.TorIpChanger._get_response_text")
    def test_real_ip(self, mock_get_response_text, mock_get):
        """
        Test that 'real_ip' is obtained without routing the request through
        Tor/Privoxy.
        """
        mock_response = Mock()
        mock_get.return_value = mock_response

        tor_ip_changer = TorIpChanger()
        tor_ip_changer.real_ip

        mock_get.assert_called_once_with(ICANHAZIP)
        mock_get_response_text.assert_called_once_with(mock_response)

    @patch("toripchanger.changer.get")
    @patch("toripchanger.changer.TorIpChanger._get_response_text")
    def test_current_ip(self, mock_get_response_text, mock_get):
        """
        Test that 'real_ip' is obtained routing the request through
        Tor/Privoxy.
        """
        mock_get_response_text.return_value = "9.9.9.9"

        tor_ip_changer = TorIpChanger()
        current_ip = tor_ip_changer.get_current_ip()

        self.assertEqual(current_ip, mock_get_response_text.return_value)

        mock_get.assert_called_once_with(
            ICANHAZIP, proxies={"http": LOCAL_HTTP_PROXY}
        )

    @patch("toripchanger.changer.get")
    def test_current_ip_exception(self, mock_get):
        """
        Test that 'get_current_ip' raises TorIpError when an IP isn't
        returned.
        """
        mock_response = Mock()
        mock_response.ok = False
        mock_get.return_value = mock_response

        tor_ip_changer = TorIpChanger()

        with self.assertRaises(TorIpError):
            tor_ip_changer.get_current_ip()

    @patch("toripchanger.changer.TorIpChanger.get_current_ip")
    @patch("toripchanger.changer.TorIpChanger._obtain_new_ip")
    def test_get_new_ip_current_ip_failed(
        self, mock_obtain_new_ip, mock_get_current_ip
    ):
        """
        Test that 'get_new_ip' attempts to get a new IP only the limited
        number of times when 'get_current_ip' keeps failing.
        """
        mock_get_current_ip.side_effect = TorIpError

        tor_ip_changer = TorIpChanger()

        with self.assertRaises(TorIpError):
            tor_ip_changer.get_new_ip()

        self.assertEqual(mock_obtain_new_ip.call_count, NEW_IP_MAX_ATTEMPTS)
        self.assertEqual(mock_get_current_ip.call_count, NEW_IP_MAX_ATTEMPTS)

    @patch("toripchanger.changer.TorIpChanger.get_current_ip")
    @patch("toripchanger.changer.TorIpChanger._obtain_new_ip")
    @patch("toripchanger.changer.TorIpChanger._ip_is_usable")
    def test_get_new_ip_ip_is_usable_failed(
        self, mock_ip_is_usable, mock_obtain_new_ip, mock_get_current_ip
    ):
        """
        Test that 'get_new_ip' attempts to get a new IP only the limited
        number of times when '_ip_is_usable' keeps returning the same IP.
        """
        mock_get_current_ip.return_value = "1.1.1.1"
        mock_ip_is_usable.return_value = False

        tor_ip_changer = TorIpChanger()
        tor_ip_changer.used_ips = ["1.1.1.1"]

        with self.assertRaises(TorIpError):
            tor_ip_changer.get_new_ip()

        self.assertEqual(mock_obtain_new_ip.call_count, NEW_IP_MAX_ATTEMPTS)
        self.assertEqual(mock_get_current_ip.call_count, NEW_IP_MAX_ATTEMPTS)
        self.assertEqual(mock_ip_is_usable.call_count, NEW_IP_MAX_ATTEMPTS)

    @patch("toripchanger.changer.TorIpChanger.get_current_ip")
    @patch("toripchanger.changer.TorIpChanger._obtain_new_ip")
    def test_get_new_ip_success(self, mock_obtain_new_ip, mock_get_current_ip):
        """
        Test that 'get_new_ip' gets a new usable Tor IP address on the third
        attempt.
        """
        mock_get_current_ip.side_effect = ["1.1.1.1", None, "2.2.2.2"]

        tor_ip_changer = TorIpChanger()
        tor_ip_changer._real_ip = "0.0.0.0"
        tor_ip_changer.used_ips = ["1.1.1.1"]

        new_ip = tor_ip_changer.get_new_ip()

        self.assertEqual(new_ip, "2.2.2.2")
        self.assertEqual([new_ip], tor_ip_changer.used_ips)
        self.assertEqual(mock_obtain_new_ip.call_count, 2)
        self.assertEqual(mock_get_current_ip.call_count, 3)

    def test_get_response_text(self):
        """
        Test that '_get_response_text' accesses and strips response's text
        property.
        """
        mock_response = Mock()
        mock_response.text = " 8.8.8.8\n"

        tor_ip_changer = TorIpChanger()
        response = tor_ip_changer._get_response_text(mock_response)

        self.assertEqual(response, "8.8.8.8")

    def test_ip_was_used(self):
        """
        Test '_ip_was_used' recognizes an already used IP.
        """
        tor_ip_changer = TorIpChanger()

        # Using '8.8.8.8' the first time.
        ip_is_safe = tor_ip_changer._ip_is_safe("8.8.8.8")
        self.assertTrue(ip_is_safe)

        # Using '8.8.8.8' again while it's still known to be used already.
        tor_ip_changer.used_ips.append("8.8.8.8")
        ip_is_safe = tor_ip_changer._ip_is_safe("8.8.8.8")
        self.assertFalse(ip_is_safe)

    def test_ip_is_usable_invalid_ip(self):
        """
        Test that '_ip_is_usable' returns False for an invalid IP.
        """
        tor_ip_changer = TorIpChanger()

        ip_usable = tor_ip_changer._ip_is_usable("not-an-ip")
        self.assertFalse(ip_usable)

    def test_ip_is_usable_real_ip(self):
        """
        Test that '_ip_is_usable' returns False for the actual real IP.
        """
        tor_ip_changer = TorIpChanger()
        tor_ip_changer._real_ip = "0.0.0.0"

        ip_usable = tor_ip_changer._ip_is_usable("0.0.0.0")
        self.assertFalse(ip_usable)

    def test_ip_is_usable_used_ip(self):
        """
        Test that '_ip_is_usable' returns False for an already used IP.
        """
        tor_ip_changer = TorIpChanger()
        tor_ip_changer._real_ip = "0.0.0.0"
        tor_ip_changer.used_ips = ["1.1.1.1"]

        ip_usable = tor_ip_changer._ip_is_usable("1.1.1.1")
        self.assertFalse(ip_usable)

    def test_ip_is_usable_valid_ip(self):
        """
        Test that '_ip_is_usable' returns True for a valid IP.
        """
        tor_ip_changer = TorIpChanger()
        tor_ip_changer._real_ip = "0.0.0.0"

        ip_usable = tor_ip_changer._ip_is_usable("1.1.1.1")
        self.assertTrue(ip_usable)

    def test_manage_used_ips_registers_ip(self):
        """
        Test that '_manage_used_ips' successfully registers current IP.
        """
        tor_ip_changer = TorIpChanger()

        current_ip = "1.1.1.1"
        tor_ip_changer._manage_used_ips(current_ip)

        self.assertEqual([current_ip], tor_ip_changer.used_ips)

    def test_manage_usable_ips_releases_used_ip(self):
        """
        Test that '_manage_used_ips' successfully releases an used IP.
        """
        tor_ip_changer = TorIpChanger()
        tor_ip_changer.used_ips = ["1.1.1.1"]

        current_ip = "2.2.2.2"
        tor_ip_changer._manage_used_ips(current_ip)

        self.assertEqual([current_ip], tor_ip_changer.used_ips)

    def test_manage_used_ips_releases_oldest_used_ip(self):
        """
        Test that '_manage_used_ips' successfully releases oldest used IP.
        """
        tor_ip_changer = TorIpChanger(3)
        tor_ip_changer._real_ip = "0.0.0.0"
        tor_ip_changer.used_ips = ["1.1.1.1", "2.2.2.2", "3.3.3.3"]

        current_ip = "4.4.4.4"
        expected_used_ips = ["2.2.2.2", "3.3.3.3", current_ip]

        tor_ip_changer._manage_used_ips(current_ip)
        self.assertEqual(tor_ip_changer.used_ips, expected_used_ips)

    @patch("toripchanger.changer.sleep")
    @patch("toripchanger.changer.Controller.from_port")
    def test_obtain_new_ip(self, mock_from_port, mock_sleep):
        """
        Test that '_obtain_new_ip' obtains new Tor IP and expected methods are
        called while doing so within the context manager.
        """
        tor_ip_changer = TorIpChanger(post_new_ip_sleep=1.0)
        tor_ip_changer._obtain_new_ip()

        mock_from_port.assert_any_call(address=TOR_ADDRESS, port=TOR_PORT)

        mock_controler = mock_from_port.return_value.__enter__()
        mock_controler.signal.assert_any_call(Signal.NEWNYM)
        mock_controler.authenticate.assert_any_call(password=TOR_PASSWORD)

        mock_sleep.assert_called_once_with(1.0)



import flask


def create_changeip_response(tor_ip_changer):
    new_ip = ""
    error = ""
    status = 200

    try:
        new_ip = tor_ip_changer.get_new_ip()
    except Exception as exc:
        error = f"{exc.__class__}: {str(exc)}"
        status = 500

    response = flask.jsonify({"newIp": new_ip, "error": error})
    response.status_code = status

    return response, status


def init_server(tor_ip_changer):
    app = flask.Flask(__name__)

    @app.route("/")
    def index():
        return "TorIpChanger Server"

    @app.route("/changeip/")
    def changeip():
        return create_changeip_response(tor_ip_changer)

    return app



"""TorIpChanger package installer"""


from os import path
from setuptools import setup


def read(fname, readlines=False):
    with open(path.join(path.abspath(path.dirname(__file__)), fname)) as f:
        return f.readlines() if readlines else f.read()


requirements = read("requirements.txt", True)
requirements_server = read("requirements-server.txt", True)

setup(
    version="1.1.1",
    name="toripchanger",
    url="https://github.com/DusanMadar/TorIpChanger",
    author="Dusan Madar",
    author_email="madar.dusan@gmail.com",
    description="Python powered way to get a unique Tor IP",
    long_description=read("README.md"),
    long_description_content_type="text/markdown",
    keywords="change tor ip",
    packages=["toripchanger", "tests"],
    scripts=["scripts/toripchanger_server"],
    include_package_data=True,
    test_suite="tests",
    license="MIT",
    platforms="linux",
    python_requires=">=3.6",
    install_requires=requirements,
    tests_require=requirements + requirements_server,
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "Programming Language :: Python :: 3",
        "Operating System :: POSIX :: Linux",
        "Natural Language :: English",
    ],
    extras_require={"server": requirements_server},
)
